import logging
from typing import Dict, List

from aporia.consts import LOGGER_NAME
from aporia.errors import AporiaError
from aporia.graphql_client import GraphQLClient


logger = logging.getLogger(LOGGER_NAME)

VALID_FIELD_TYPES = ["numeric", "categorical", "string", "boolean", "datetime"]


async def run_create_model_version_query(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    features: Dict[str, str],
    prediction: Dict[str, str],
):
    """Defines the schema for a specific model version.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model Version
        features (Dict[str, str]): Feature fields
        prediction (Dict[str, str]): Prediction fields
    """
    query = """
        mutation CreateModelVersion(
            $modelId: String!,
            $modelVersion: String!,
            $features: [Field]!,
            $prediction: [Field]!
        ) {
            createModelVersion(
                modelId: $modelId,
                modelVersion: $modelVersion,
                features: $features,
                prediction: $prediction
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "features": prepare_fields(features),
        "prediction": prepare_fields(prediction),
    }

    result = await graphql_client.query_with_retries(query, variables)
    for warning in result["createModelVersion"]["warnings"]:
        logger.warning(warning)


def prepare_fields(fields: Dict[str, str]) -> List[Dict[str, str]]:
    """Creates a list of Field GraphQL objects from a fields dict.

    Args:
        fields (Dict[str, str]): Fields dict

    Returns:
        List[Dict[str, str]]: List of GraphQL Field objects.
    """
    return [{"name": field_name, "type": field_type} for field_name, field_type in fields.items()]


def validate_fields_input(features: Dict[str, str], prediction: Dict[str, str]):
    """Checks if the fields passed to create_model_version are valid.

    Args:
        features (Dict[str, str]): Features fields
        prediction (Dict[str, str]): Prediction fields
    """
    for category, fields in [("features", features), ("prediction", prediction)]:
        if not isinstance(fields, dict):
            raise AporiaError("{} parameter must be a dict".format(category))

        if len(fields) == 0:
            raise AporiaError("{} parameter must contain items".format(category))

        for key, value in fields.items():
            if not isinstance(key, str):
                raise AporiaError(
                    "Invalid field name {} in the {} parameter - field names must be strings".format(
                        key, category
                    )
                )

            if value not in VALID_FIELD_TYPES:
                raise AporiaError(
                    "Invalid field type {} in the {} parameter - valid field types are {}".format(
                        value, category, VALID_FIELD_TYPES
                    )
                )



import asyncio
from concurrent.futures import Future
import logging
from threading import Thread
from typing import Any, Coroutine

from aporia.consts import LOGGER_NAME


logger = logging.getLogger(LOGGER_NAME)


class EventLoop:
    """Asyncio event loop manager for worker thread loop."""

    def __init__(self):
        """Initializes an EventLoop instance."""
        logger.debug("Starting event loop in a worker thread.")
        self.loop = asyncio.new_event_loop()
        self.loop_thread = Thread(target=self.run_loop, daemon=True)

        self.loop_thread.start()

    def run_loop(self):
        """Runs the asyncio event loop."""
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def create_task(self, coro: Coroutine) -> Future:
        """Schedule a coroutine without waiting for result.

        Args:
            coro (Coroutine): Coroutine to schedule.

        Returns:
            Future: Future object to wait for the result of the coroutine.
        """
        future = asyncio.run_coroutine_threadsafe(coro=coro, loop=self.loop)
        return future

    def run_coroutine(self, coro: Coroutine) -> Any:
        """Runs a coroutine in the worker thread loop and await result.

        Args:
            coro (Coroutine): Coroutine to execute

        Returns:
            Any: Coroutine return value
        """
        future = self.create_task(coro)

        return future.result()



from concurrent.futures import ALL_COMPLETED, wait
from datetime import datetime, timezone
import logging
from typing import Any, cast, Dict, Iterable, List, Optional, Union

import aporia
from aporia.api.log_json import log_json
from aporia.api.log_predict import (
    is_valid_prediction,
    log_batch_prediction,
    serialize_prediction_batch,
)
from aporia.config import Config
from aporia.consts import LOGGER_NAME
from aporia.errors import handle_error
from aporia.event_loop import EventLoop
from aporia.graphql_client import GraphQLClient
from aporia.prediction_queue import PredictionQueue


logger = logging.getLogger(LOGGER_NAME)


SupportedValueType = Union[float, int, bool, str]


class BaseModel:
    """Base class for Model objects."""

    def __init__(
        self,
        model_id: str,
        model_version: str,
        graphql_client: Optional[GraphQLClient],
        event_loop: Optional[EventLoop],
        config: Optional[Config],
    ):
        """Initializes a BaseModel object.

        Args:
            model_id (str): Model identifier
            model_version (str): Model version
            graphql_client (Optional[GraphQLClient]): GraphQL client
            event_loop (Optional[EventLoop]): AsyncIO event loop
            config (Optional[Config]): Aporia config
        """
        logger.debug(
            "Initializing model object for model {} version {}".format(model_id, model_version)
        )
        self._model_ready = False

        self.model_id = model_id
        self.model_version = model_version
        self._graphql_client = cast(GraphQLClient, graphql_client)
        self._event_loop = cast(EventLoop, event_loop)
        self._config = cast(Config, config)
        # We keep a list of all tasks that were not awaited, to allow flushing
        # We have to do this manually to support python versions below
        # 3.7 (otherwise we could use asyncio.all_tasks())
        self._futures = []  # type: ignore

        if event_loop is not None and graphql_client is not None and config is not None:
            prediction_queue = self._init_prediction_queue()
            if prediction_queue is not None:
                self._prediction_queue = prediction_queue
                self._model_ready = True

        if not self._model_ready:
            logger.error("Model object initialization failed - model operations will not work.")

    def _init_prediction_queue(self) -> Optional[PredictionQueue]:
        try:
            return PredictionQueue(
                event_loop=self._event_loop,
                batch_size=self._config.prediction_queue_batch_size,
                max_size=self._config.prediction_queue_max_predictions,
                flush_interval=self._config.prediction_queue_flush_interval,
                flush_callback=self._log_batch_prediction,
            )
        except Exception as err:
            handle_error(
                message="Initializing prediction queue failed, error: {}".format(str(err)),
                add_trace=self._config.debug,
                raise_exception=self._config.throw_errors,
                original_exception=err,
                log_level=logging.ERROR,
            )
            return None

    def log_prediction(
        self,
        id: str,
        features: Dict[str, SupportedValueType],
        prediction: Dict[str, SupportedValueType],
        occurred_at: Optional[datetime] = None,
    ):
        """Logs a single prediction.

        Args:
            id (str): Prediction identifier.
            features (Dict[str, SupportedValueType]): Values for all the features in the prediction
            prediction (Dict[str, SupportedValueType]): Prediction result
            occurred_at (datetime, optional): Prediction timestamp. Defaults to None.

        Note:
            * If occurred_at is None, it will be reported as datetime.now()
        """
        self.log_batch_prediction(
            predictions=[
                dict(
                    id=id,
                    features=features,
                    prediction=prediction,
                    occurred_at=occurred_at,
                )
            ]
        )

    def log_batch_prediction(self, predictions: Iterable[dict]):
        """Logs multiple predictions.

        Args:
            predictions (Iterable[dict]): An iterable that produces prediction dicts.
                Each prediction dict MUST contain the following keys:
                    * id (str): Prediction identifier.
                    * features (Dict[str, SupportedValueType]): Values for all the features
                        in the prediction
                    * prediction (Dict[str, SupportedValueType]): Prediction result
                Each prediction dict MAY also contain the following keys:
                    * occurred_at (datetime): Prediction timestamp.

        Notes:
            * If occurred_at is None in any of the predictions, it will be reported as datetime.now()
        """
        if not self._model_ready:
            return

        logger.debug("Logging predictions")
        try:
            if self._config.debug:
                logger.debug("Validating predictions")
                for i, predict in enumerate(predictions):
                    if is_valid_prediction(predict):
                        logger.debug("Prediction {} is valid".format(i))
                    else:
                        logger.debug("Prediction {} is not valid".format(i))

            now = datetime.now(tz=timezone.utc)
            predictions_with_timestamps = [
                {"data": predict, "timestamp": now} for predict in predictions
            ]

            prediction_count = self._event_loop.run_coroutine(
                self._prediction_queue.put(predictions=predictions_with_timestamps)
            )

            dropped_predictions = len(predictions_with_timestamps) - prediction_count
            if dropped_predictions > 0:
                logger.warning(
                    "Reached prediction limit, dropped {} predictions.".format(dropped_predictions)
                )

        except Exception as err:
            handle_error(
                message="Logging prediction batch failed, error: {}".format(str(err)),
                add_trace=self._config.debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    async def _log_batch_prediction(self, predictions_with_timestamps: List[dict]):
        try:
            serialized_predictions = serialize_prediction_batch(predictions_with_timestamps)

            if len(serialized_predictions) > 0:
                await log_batch_prediction(
                    graphql_client=self._graphql_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    environment=self._config.environment,
                    serialized_predictions=serialized_predictions,
                    await_insert=self._config.debug,
                )
        except Exception as err:
            handle_error(
                message="Logging prediction failed, error: {}".format(str(err)),
                add_trace=self._config.debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    def log_json(self, data: Any):
        """Logs arbitrary data.

        Args:
            data (Any): Data to log, must be JSON serializable
        """
        if not self._model_ready:
            return

        logger.debug("Logging arbitrary data.")
        try:
            future = self._event_loop.create_task(self._log_json(data=data))
            self._futures.append(future)
        except Exception as err:
            handle_error(
                message="Logging arbitrary data failed, error: {}".format(str(err)),
                add_trace=self._config.debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    async def _log_json(self, data: Any):
        try:
            await log_json(
                graphql_client=self._graphql_client,
                model_id=self.model_id,
                model_version=self.model_version,
                environment=self._config.environment,
                data=data,
            )
        except Exception as err:
            handle_error(
                message="Logging arbitrary data failed, error: {}".format(str(err)),
                add_trace=self._config.debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    def flush(self, timeout: Optional[int] = None) -> Optional[int]:
        """Waits for all currently scheduled tasks to finish.

        Args:
            timeout (int, optional): Maximum number of seconds to wait for tasks to
                complete. Default to None (No timeout).

        Returns:
            Optional[int]: Number of tasks that haven't finished running.
        """
        if not self._model_ready:
            return None

        try:
            futures = self._futures
            self._futures = []

            logger.debug("Flusing predictions.")
            # Add a task that flushes the queue, and another that waits for the flush to complete
            futures.append(self._event_loop.create_task(self._prediction_queue.flush()))
            futures.append(self._event_loop.create_task(self._prediction_queue.join()))

            logger.debug("Waiting for {} scheduled tasks to finish executing.".format(len(futures)))
            done, not_done = wait(futures, timeout=timeout, return_when=ALL_COMPLETED)

            logger.debug(
                "{} tasks finished, {} tasks not finished.".format(len(done), len(not_done))
            )
            self._futures.extend(not_done)

            return len(not_done)
        except Exception as err:
            handle_error(
                message="Flushing remaining data failed, error: {}".format(str(err)),
                add_trace=self._config.debug,
                raise_exception=self._config.throw_errors,
                original_exception=err,
                log_level=logging.ERROR,
            )

        return None


class Model(BaseModel):
    """Model object."""

    def __init__(self, model_id: str, model_version: str):
        """Initializes a Model object.

        Args:
            model_id (str): Model identifier, as received from the Aporia dashboard.
            model_version (str): Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
        """
        if aporia.context is None:
            logger.error("Aporia was not initialized.")
            super().__init__(
                model_id=model_id,
                model_version=model_version,
                graphql_client=None,
                event_loop=None,
                config=None,
            )
        else:
            super().__init__(
                model_id=model_id,
                model_version=model_version,
                graphql_client=aporia.context.graphql_client,
                event_loop=aporia.context.event_loop,
                config=aporia.context.config,
            )



from datetime import datetime
import logging
from typing import List

from aporia.consts import LOGGER_NAME
from aporia.errors import AporiaError
from aporia.graphql_client import GraphQLClient

logger = logging.getLogger(LOGGER_NAME)


REQUIRED_FIELDS = {"id", "features", "prediction"}


async def log_batch_prediction(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    environment: str,
    serialized_predictions: List[dict],
    await_insert: bool,
):
    """Reports a batch of predictions.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        environment (str): Environment in which aporia is running.
        serialized_predictions (List[dict]): List of serialized prediction dicts.
        await_insert (bool): True if the controller should wait for the predictions
            to be stored before responding to the sdk.
    """
    query = """
        mutation LogPredict(
            $modelId: String!,
            $modelVersion: String!,
            $environment: String!,
            $predictions: [Prediction]!
            $isSync: Boolean!
        ) {
            logPredictions(
                modelId: $modelId,
                modelVersion: $modelVersion,
                environment: $environment,
                predictions: $predictions
                isSync: $isSync
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "environment": environment,
        "predictions": serialized_predictions,
        "isSync": await_insert,
    }

    result = await graphql_client.query_with_retries(query, variables)
    for warning in result["logPredictions"]["warnings"]:
        logger.warning(warning)


def serialize_prediction_batch(predictions: List[dict]) -> List[dict]:
    """Serializes prediction batches.

    Args:
        predictions (List[dict]): Prediction dicts with user data and timestamp

    Returns:
        List[dict]: Serialized prediction batch.
    """
    prepared_predictions = []  # type: ignore

    for predict in predictions:
        try:
            predict_data = predict["data"]

            missing_fields = REQUIRED_FIELDS - predict_data.keys()
            if len(missing_fields) > 0:
                raise AporiaError("Missing required fields {}".format(missing_fields))

            occurred_at = predict_data.get("occurred_at")
            if occurred_at is None:
                occurred_at = predict["timestamp"]

            prepared_predict = {
                "id": predict_data["id"],
                "features": predict_data["features"],
                "prediction": predict_data["prediction"],
                "occurredAt": occurred_at.isoformat(),
            }

            prepared_predictions.append(prepared_predict)

        except Exception as err:
            logger.error("Failed to serialize prediction, error: {}".format(err))
            continue

    return prepared_predictions


def is_valid_prediction(predict: dict) -> bool:
    """Checks if a prediction is valid.

    This function exists mostly for debugging purposes.

    Args:
        predict (dict): Prediction dict

    Returns:
        bool: True if the prediction is valid.

    Notes:
        * This function is only used in debug mode.
    """
    missing_fields = REQUIRED_FIELDS - predict.keys()
    if len(missing_fields) > 0:
        logger.debug("Invalid prediction, missing required fields {}".format(missing_fields))
        return False

    if not (isinstance(predict["features"], dict) and len(predict["features"]) > 0):
        logger.debug("Invalid input format for features parameter - expected a non-empty dict")
        return False

    if not (isinstance(predict["prediction"], dict) and len(predict["prediction"]) > 0):
        logger.debug("Invalid input format for prediction parameter - expected a non-empty dict")
        return False

    if predict.get("occurred_at") is not None and not isinstance(predict["occurred_at"], datetime):
        logger.debug("Invalid input format for occurred_at parameter - expected datetime")
        return False

    return True



import atexit
from contextlib import suppress
import logging
import sys
from typing import Dict, Optional

from aporia.api.create_model_version import run_create_model_version_query, validate_fields_input
from aporia.config import Config
from aporia.consts import LOG_FORMAT, LOGGER_NAME
from aporia.errors import AporiaError, handle_error
from aporia.event_loop import EventLoop
from aporia.graphql_client import GraphQLClient
from aporia.model import Model

try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"


__all__ = ["init", "shutdown", "Model"]


class Context:
    """Global context."""

    def __init__(
        self,
        graphql_client: GraphQLClient,
        event_loop: EventLoop,
        config: Config,
    ):
        """Initializes the context.

        Args:
            graphql_client (GraphQLClient): GraphQL client.
            event_loop (EventLoop): Event loop.
            config (Config): Aporia config.
        """
        self.graphql_client = graphql_client
        self.event_loop = event_loop
        self.config = config


context = None
logger = logging.getLogger(LOGGER_NAME)


def init(
    token: Optional[str] = None,
    host: Optional[str] = None,
    environment: Optional[str] = None,
    port: Optional[int] = None,
    debug: Optional[bool] = None,
    throw_errors: Optional[bool] = None,
    blocking_call_timeout: Optional[int] = None,
):
    """Initializes the Aporia SDK.

    Args:
        token (str): Authentication token.
        host (str): Controller host.
        environment (str): Environment in which aporia is initialized (e.g production, staging).
        port (int, optional): Controller port. Defaults to 443.
        debug (bool, optional): True to enable debug mode - this will cause additional logs
            and stack traces during exceptions. Defaults to False.
        throw_errors (bool, optional): True to cause errors to be raised
            as exceptions. Defaults to False.
        blocking_call_timeout (int, optional): Timeout, in seconds, for blocking aporia API
            calls - Model(), set_features(), add_extra_inputs(), add_extra_outputs().

    Notes:
        * The token, host and environment parameters are required.
        * All of the parameters here can also be defined as environment variables:
            * token -> APORIA_TOKEN
            * host -> APORIA_HOST
            * environment -> APORIA_ENVIRONMENT
            * port -> APORIA_PORT
            * debug -> APORIA_DEBUG
            * throw_errors -> APORIA_THROW_ERRORS
            * blocking_call_timeout -> APORIA_BLOCKING_CALL_TIMEOUT
        * Values passed as parameters to aporia.init() override the values from
          the corresponding environment variables.
    """
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if not logger.hasHandlers():
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT, style="{"))
        logger.addHandler(handler)

    logger.debug("Initializing Aporia SDK.")

    try:
        config = Config(
            token=token,
            host=host,
            environment=environment,
            port=port,
            debug=debug,
            throw_errors=throw_errors,
            blocking_call_timeout=blocking_call_timeout,
        )

        # Init graphql client and event loop
        event_loop = EventLoop()
        graphql_client = GraphQLClient(token=config.token, host=config.host, port=config.port)
        event_loop.run_coroutine(graphql_client.open())

        global context
        context = Context(
            graphql_client=graphql_client,
            event_loop=event_loop,
            config=config,
        )

        atexit.register(shutdown)
        logger.debug("Aporia SDK initialized.")
    except Exception as err:
        handle_error(
            message="Initializing Aporia SDK failed, error: {}".format(str(err)),
            add_trace=False if debug is None else debug,
            raise_exception=False if throw_errors is None else throw_errors,
            original_exception=err,
        )


def shutdown():
    """Shuts down the Aporia SDK.

    Notes:
        * It is advised to call flush() before calling shutdown(), to ensure that
          all of the data that was sent reaches the controller.
    """
    with suppress(Exception):
        global context
        if context is not None:
            context.event_loop.run_coroutine(context.graphql_client.close())
            context = None


def create_model_version(
    model_id: str, model_version: str, features: Dict[str, str], prediction: Dict[str, str]
):
    """Create a new model version, and defines a schema for it.

    Args:
        model_id (str): Model identifier, as received from the Aporia dashboard.
        model_version (str): Model version - this can be any string that represents the model
            version, such as "v1" or a git commit hash.
        features (Dict[str, str]): Schema for model features (See notes)
        prediction (Dict[str, str]): Schema for prediction results (See notes)

    Notes:
        * A schema is a dict, in which the keys are the fields you wish to report, and the
          values are the types of those fields. For example:
            {
                "feature1": "numeric",
                "feature2": "datetime"
            }
        * The valid field types (and corresponding python types) are:
            | Field Type    | Python Types
            | ------------- | ------------
            | "numeric"     | float, int
            | "categorical" | int
            | "boolean"     | bool
            | "string"      | str
            | "datetime"    | datetime.datetime, or str representing a datetime in ISO-8601 format
    """
    if context is None:
        raise AporiaError("Aporia SDK was not initialized")

    try:
        validate_fields_input(features=features, prediction=prediction)

        context.event_loop.run_coroutine(
            run_create_model_version_query(
                graphql_client=context.graphql_client,
                model_id=model_id,
                model_version=model_version,
                features=features,
                prediction=prediction,
            )
        )
    except Exception as err:
        handle_error(
            message="Creating model version failed, error: {}".format(str(err)),
            add_trace=context.config.debug,
            raise_exception=context.config.throw_errors,
            original_exception=err,
        )



import logging
from typing import Optional

from aporia.consts import LOGGER_NAME


logger = logging.getLogger(LOGGER_NAME)


class AporiaError(Exception):
    """Base class for Aporia SDK exceptions."""

    pass


def handle_error(
    message: str,
    add_trace: bool,
    raise_exception: bool,
    log_level: int = logging.ERROR,
    original_exception: Optional[Exception] = None,
):
    """Handles an error with either a log or an exception.

    Args:
        message (str): Error message
        add_trace (bool): True if stack trace should be added to the log or exception
        raise_exception (bool): True if exceptions should be raised
        log_level (int): Log level for log messages. Defaults to logging.ERROR.
        original_exception (Exception, optional): Original exception for stack trace. Defaults to None.
    """
    if raise_exception:
        if add_trace:
            raise AporiaError(message) from original_exception
        else:
            raise AporiaError(message) from None
    else:
        logger.log(level=log_level, msg=message, exc_info=add_trace)



import asyncio
import logging
from typing import Awaitable, Callable, Iterable, List

from aporia.consts import LOGGER_NAME
from aporia.event_loop import EventLoop


logger = logging.getLogger(LOGGER_NAME)


class PredictionQueue:
    """Queue for prediction batching."""

    def __init__(
        self,
        event_loop: EventLoop,
        batch_size: int,
        max_size: int,
        flush_interval: int,
        flush_callback: Callable[[List[dict]], Awaitable[None]],
    ):
        """Initializes a PredictionQueue.

        Args:
            event_loop (EventLoop): Event loop for async operations
            batch_size (int): Prediction batch size
            max_size (int): Maximum number of predictions in the queue.
            flush_interval (int): Interval for periodic flushes.
            flush_callback (Callable[[List[dict]], Awaitable[None]]): An awaitable function
                to call to flush predictions.
        """
        self.batch_size = batch_size
        self.max_size = max_size
        self.partial_batch = []  # type: ignore
        self.scheduled_batches = 0
        self.flush_callback = flush_callback

        # The queue has to be initialized in the asyncio loop
        self.queue = event_loop.run_coroutine(self._init_queue())
        event_loop.create_task(self._flush_on_interval(flush_interval))

    @staticmethod
    async def _init_queue() -> asyncio.Queue:
        return asyncio.Queue()

    def __len__(self) -> int:
        """Calculates the length of the queue.

        Returns:
            int: Total number of predictions in the queue.
        """
        # Since we want to limit the number of predictions in memory, we have to consider not
        # only the batches that are in the queue, but also the ones that have been scheduled
        # for sending but have not yet been sent.
        return (
            self.scheduled_batches * self.batch_size
            + len(self.partial_batch)
            + self.queue.qsize() * self.batch_size
        )

    async def _flush_on_interval(self, interval: int):
        while True:
            await asyncio.sleep(interval)
            await self.flush()

    async def put(self, predictions: Iterable[dict]) -> int:
        """Inserts predictions to the queue.

        Args:
            predictions (Iterable[dict]): Predictions to insert

        Returns:
            int: Number of predictions added to the queue.

        Notes;
            * Any predictions above the max queue size will be discarded.
        """
        for count, predict in enumerate(predictions, start=1):
            if len(self) >= self.max_size:
                logger.warning(
                    "Maximum number of predictions reached, dropping remaining predictions."
                )
                count -= 1
                break

            if len(self.partial_batch) < self.batch_size:
                self.partial_batch.append(predict)

            if len(self.partial_batch) == self.batch_size:
                # The queue is only accessed from the asyncio loop, and it has no
                # max size, to put_nowait is safe.
                self.queue.put_nowait(self.partial_batch)
                self.partial_batch = []

        if not self.queue.empty():
            asyncio.ensure_future(self.flush())

        return count

    async def flush(self):
        """Flushes the prediction queue."""
        if len(self.partial_batch) > 0:
            self.queue.put_nowait(self.partial_batch)
            self.partial_batch = []

        flush_tasks = []
        # The queue is only accessed from the asyncio loop, meaning only
        # one flush() will run at any given moment, so get_nowait is safe here.
        while not self.queue.empty():
            batch = self.queue.get_nowait()
            flush_tasks.append(self.flush_callback(batch))

        self.scheduled_batches += len(flush_tasks)
        await asyncio.gather(*flush_tasks, return_exceptions=True)

        self.scheduled_batches -= len(flush_tasks)
        for _ in flush_tasks:
            self.queue.task_done()

    async def join(self):
        """Wait for all scheduled predictions to be sent."""
        await self.queue.join()
